import abc
import json
import pathlib
import typing
import uuid

import yaml
from kubernetes import client

PathLike = typing.Union[str, pathlib.Path]
OptionalPathLike = typing.Optional[PathLike]

API_VERSION_REMAPS = {"core/v1": "v1"}


# Protocol was added in 3.8 and needs to be gracefully handled for now
# without the Protocol class for static duck-typing. Eventually this
# should be replaced by:
# class ExecutionResponse(typing.Protocol):
class ExecutionResponse:
    """Structural type for kubernetes execution responses."""

    @property
    def status(self) -> "Definition":
        """Status from an execution response."""
        pass


class InternalValue(typing.NamedTuple):
    """
    Data structure that can be used for internal value states that help manage
    the state of user input versus internally set values.
    """

    #: Unique identifier for the internal value.
    name: str
    #: A brief description for the purpose of the internal value.
    description: str


#: Uniquely identifying value that can be used to represent a state where the
#: user has not modified the value. Useful in cases where `None` could
#: represent a significant change.
UNCHANGED_VALUE = InternalValue(
    name="UnchangedValue", description="No change is being made to this value."
)


class Definition:
    """Base class for all kubernetes definition objects."""

    def __init__(self, api_version: str, kind: str):
        """Initialize basic common properties"""
        self._api_version = api_version
        self._kind = kind
        self._properties: typing.Dict[str, typing.Any] = {}
        self._types: typing.Dict[str, typing.Any] = {}
        self._kuber_uid = str(uuid.uuid4())

    @property
    def kuber_uid(self) -> str:
        """Internal UID used to uniquely identify this object."""
        return self._kuber_uid

    def to_dict(self) -> typing.Optional[dict]:
        """
        Serializes the `Definition` object to a dictionary that can be
        rendered into json or yaml configuration file formats.
        """
        results = {
            p: serialize_property(v)
            for p, v in self._properties.items()
            if p in self._types
        }
        results = {
            p: v for p, v in results.items() if v or isinstance(v, bool) or v == 0
        }
        return results if results else None

    def from_dict(self, source: dict) -> "Definition":
        """Populates the resource from the source dictionary definition."""
        lower_keys = {k.lower(): k for k in self._types.keys()}
        for key, value in source.items():
            camel_key = to_camel_case(key)
            if key in self._types:
                prop_key = key
            elif camel_key in self._types:
                prop_key = camel_key
            elif camel_key.lower() in lower_keys:
                prop_key = lower_keys[camel_key.lower()]
            else:  # pragma: no cover
                continue

            self._properties[prop_key] = deserialize_property(
                value=value, data_type=self._types[prop_key]
            )

        return self


class Collection(Definition):
    """
    Base class for all kubernetes collection resources, which are synthetic
    resources that bundle together multiple standard resources.
    """

    def __init__(self, api_version: str, kind: str):
        """Initialize basic common properties."""
        super(Collection, self).__init__(api_version, kind)

    @property
    def metadata(self):
        """Must be implemented by subclasses"""
        return

    @metadata.setter
    def metadata(self, value):
        """Must be implemented by subclasses"""
        pass

    @property
    def api_version(self) -> str:
        """The Kubernetes API version in which the resource resides."""
        return self._api_version

    @property
    def kind(self) -> str:
        """Resource type identifier."""
        return self._kind

    def to_dict(self) -> dict:
        """
        Serializes the `Resource` object to a dictionary that can be
        rendered into json or yaml configuration file formats.
        """
        results = super(Collection, self).to_dict() or {}
        version = API_VERSION_REMAPS.get(self.api_version, self.api_version)
        return {"apiVersion": version, "kind": self.kind, **results}

    def to_json(self) -> str:
        """
        Serializes the `Resource` object to a json string that can be
        saved as a configuration file or used with other packages that
        interact with the Kubernetes API.
        """
        return json.dumps(self.to_dict(), indent=2)

    def to_yaml(self) -> str:
        """
        Serializes the `Resource` object to a yaml string that can be
        saved as a configuration file or used with other packages that
        interact with the Kubernetes API.
        """
        return yaml.dump(self.to_dict(), default_flow_style=False)

    @staticmethod
    @abc.abstractmethod
    def get_resource_api(api_client: client.ApiClient = None, **kwargs) -> typing.Any:
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        ...  # pragma: no cover


class Resource(Definition):
    """
    Base class for all kubernetes resource definition objects. All entities
    in the Kubernetes APIs are Definitions, but Resources are the top-level
    entities that have the standard `apiVersion`, `kind`, and `metadata`
    elements.
    """

    def __init__(self, api_version: str, kind: str):
        """Initialize basic common properties."""
        super(Resource, self).__init__(api_version, kind)

    @property
    def metadata(self):
        """Must be implemented by subclasses"""
        return

    @metadata.setter
    def metadata(self, value):
        """Must be implemented by subclasses"""
        pass

    @property
    def api_version(self) -> str:
        """The Kubernetes API version in which the resource resides."""
        return self._api_version

    @property
    def kind(self) -> str:
        """Resource type identifier."""
        return self._kind

    def to_dict(self) -> dict:
        """
        Serializes the `Resource` object to a dictionary that can be
        rendered into json or yaml configuration file formats.
        """
        results = super(Resource, self).to_dict() or {}
        version = API_VERSION_REMAPS.get(self.api_version, self.api_version)
        return {"apiVersion": version, "kind": self.kind, **results}

    def to_json(self) -> str:
        """
        Serializes the `Resource` object to a json string that can be
        saved as a configuration file or used with other packages that
        interact with the Kubernetes API.
        """
        return json.dumps(self.to_dict(), indent=2)

    def to_yaml(self) -> str:
        """
        Serializes the `Resource` object to a yaml string that can be
        saved as a configuration file or used with other packages that
        interact with the Kubernetes API.
        """
        return yaml.dump(self.to_dict(), default_flow_style=False)

    @abc.abstractmethod
    def create_resource(self, namespace: str = None):
        """Must be implemented by subclasses."""
        pass  # pragma: no cover

    @abc.abstractmethod
    def replace_resource(self, namespace: "str" = None):
        """Must be implemented by subclasses."""
        pass  # pragma: no cover

    @abc.abstractmethod
    def patch_resource(self, namespace: "str" = None):
        """Must be implemented by subclasses."""
        pass  # pragma: no cover

    @abc.abstractmethod
    def get_resource_status(self, namespace: str = None):
        """Must be implemented by subclasses."""
        pass  # pragma: no cover

    @abc.abstractmethod
    def delete_resource(self, namespace: str = None):
        """Must be implemented by subclasses."""
        pass  # pragma: no cover

    @abc.abstractmethod
    def read_resource(self, namespace: str = None):
        """Must be implemented by subclasses."""
        pass  # pragma: no cover

    @staticmethod
    @abc.abstractmethod
    def get_resource_api(api_client: client.ApiClient = None, **kwargs) -> typing.Any:
        """
        Returns an instance of the kubernetes API client associated with
        this object.
        """
        pass  # pragma: no cover


def serialize_property(value: typing.Any) -> typing.Any:
    """
    Converts a Definition property value into one compatible with the json
    and yaml configuration formats. Basic types are returned without any
    conversion, but Definition objects are converted to their dictionary
    representation and lists of non-basic types are converted down as well.

    :param value:
        Property value to be serialized for configuration file format
        compatibility.
    :return:
        Serialized version of the supplied value.
    """
    # Some falsy cases are meaningful and should be preserved.
    nulls: typing.Tuple[None, dict, list] = (None, {}, [])

    if hasattr(value, "to_dict"):
        return value.to_dict() or None

    if isinstance(value, (list, tuple)):
        raw_values = [serialize_property(v) for v in value]
        return [v for v in raw_values if v not in nulls] or None

    if isinstance(value, dict):
        raw_items = {k: serialize_property(v) for k, v in value.items()}
        return {k: v for k, v in raw_items.items() if v not in nulls} or None

    return None if value in nulls else value


def deserialize_property(value: typing.Any, data_type: tuple) -> typing.Any:
    """
    Converts the serialized configuration value into the given kuber data type.
    All Definition objects store data types internally to allow for easily
    mapping from serialized values during the loading of configuration file
    data.

    :param value:
        Serialized configuration value to be converted into the given data
        type.
    :param data_type:
        kuber data type to deserialize the given value into.
    :return:
        Deserialized form of the specified value.
    """
    if value is None:
        return None

    if issubclass(data_type[0], Definition):
        return data_type[0]().from_dict(value)

    if data_type[0] == list:
        return [deserialize_property(v, (data_type[1], None)) for v in value]

    return data_type[0](value)


def to_camel_case(source: str) -> str:
    """Converts kebab-case or snake_case to camelCase."""
    parts = source.replace("-", "_").split("_")
    return "".join([parts[0], *[p.capitalize() for p in parts[1:]]])



import importlib
import json
import typing

import yaml

import kuber
from kuber.definitions import Resource

ResourceSubclass = typing.Union[Resource, typing.Any]


def from_yaml_file(
    file_path: str, kubernetes_version: "kuber.VersionLabel" = "latest"
) -> typing.Optional[ResourceSubclass]:
    """
    Creates a Resource object from a YAML configuration file.

    :param file_path:
        Path to the kubernetes resource configuration YAML file to be
        loaded.
    :param kubernetes_version:
        Version of the kubernetes API to use when creating the Resource. If
        omitted, the `latest` version will be used by default. Accepts either
        a string version label of a KubernetesVersion object.
    """
    with open(file_path) as f:
        return from_yaml(f.read(), kubernetes_version)


def from_yaml_file_multiple(
    file_path: str, kubernetes_version: "kuber.VersionLabel" = "latest"
) -> typing.List[ResourceSubclass]:
    """
    Creates Resource objects for each document found in the YAML file.
    Empty documents will be ignored.

    :param file_path:
        Path to the kubernetes resources configuration YAML file to be
        loaded.
    :param kubernetes_version:
        Version of the kubernetes API to use when creating the Resources. If
        omitted, the `latest` version will be used by default. Accepts either
        a string version label of a KubernetesVersion object.
    """
    with open(file_path) as f:
        return from_yaml_multiple(f.read(), kubernetes_version)


def from_yaml_multiple(
    resources_definitions: str, kubernetes_version: "kuber.VersionLabel" = "latest"
) -> typing.List[ResourceSubclass]:
    """
    Creates Resource objects for each document found in the YAML string.
    Empty documents will be ignored.

    :param resources_definitions:
        String containing the contents of one or more yaml resource
        definitions separated by ``\n---``.
    :param kubernetes_version:
        Version of the kubernetes API to use when creating the Resources. If
        omitted, the ``latest`` version will be used by default. Accepts
        either a string version label of a KubernetesVersion object.
    """
    resources = [
        from_yaml(d, kubernetes_version) for d in resources_definitions.split("\n---")
    ]
    return [r for r in resources if r is not None]


def from_yaml(
    resource_definition: str, kubernetes_version: "kuber.VersionLabel" = "latest"
) -> typing.Optional[ResourceSubclass]:
    """
    Creates a Resource object from a YAML string.

    :param resource_definition:
        String containing the contents of a yaml resource definition.
    :param kubernetes_version:
        Version of the kubernetes API to use when creating the Resource. If
        omitted, the `latest` version will be used by default. Accepts either
        a string version label of a KubernetesVersion object.
    """
    data = yaml.load(resource_definition, Loader=kuber.yaml_loader)
    return from_dict(data, kubernetes_version)


def new_resource(
    api_version: str,
    kind: str,
    name: str = None,
    kubernetes_version: "kuber.VersionLabel" = None,
    **kwargs: str,
) -> typing.Optional[ResourceSubclass]:
    """
    Creates an empty Kubernetes resource object of the specified type for
    the specified Kubernetes version.

    :param api_version:
        A standard Kubernetes configuration api version, e.g. "apps/v1", as
        to where to locate this resource within the Kubernetes API.
    :param kind:
        The type of resource, e.g. "Deployment". Case matches and should
        reflect the `kind` as specified in a Kubernetes configuration file.
    :param name:
        Name to give to the resource in its metadata.
    :param kubernetes_version:
        Version of the kubernetes API to use when creating the Resource. If
        omitted, the `latest` version will be used by default. Accepts either
        a string version label of a KubernetesVersion object.
    :param kwargs:
        Labels to assign to the metadata of the new resource.
    """
    definition = {
        "apiVersion": api_version,
        "kind": kind,
        "metadata": {"name": name, "labels": kwargs},
    }
    return from_dict(definition, kubernetes_version or "latest")


def from_dict(
    resource_definition: dict, kubernetes_version: "kuber.VersionLabel" = "latest"
) -> typing.Optional[ResourceSubclass]:
    """
    Converts a dictionary into a Resource object.

    :param resource_definition:
        Definition dictionary object to deserialize.
    :param kubernetes_version:
        Version of the kubernetes API to use when creating the Resource. If
        omitted, the `latest` version will be used by default. Accepts either
        a string version label of a KubernetesVersion object.
    """
    if not resource_definition:
        return None

    version: str = getattr(kubernetes_version, "label", kubernetes_version or "latest")
    if version.find(".") > 0 and not version.startswith("v"):
        version = f"v{version}"
    version = version.replace(".", "_")

    parts = (
        resource_definition["apiVersion"]
        .replace("rbac.authorization.k8s.io/", "rbac/")
        .replace("apiregistration.k8s.io/", "apiregistration/")
        .replace("storage.k8s.io/", "storage/")
        .split("/")[:2]
    )
    area = parts[-1]
    group = parts[0] if len(parts) > 1 else "core"
    package = ".".join(["kuber", f"{version}", f"{group}_{area}"])

    try:
        loaded_module = importlib.import_module(package)
    except ModuleNotFoundError as error:  # pragma: no cover
        print(
            f'Error: Unable to import module "{package}" '
            f'for resource apiVersion: "{resource_definition["apiVersion"]}".'
        )
        raise error

    resource_class = getattr(loaded_module, resource_definition["kind"])
    resource: Resource = resource_class()
    return resource.from_dict(resource_definition)


def from_json_file(
    file_path: str, kubernetes_version: "kuber.VersionLabel" = "latest"
) -> typing.Optional[ResourceSubclass]:
    """
    Creates a Resource object from a configuration file.

    :param file_path:
        Path to a JSON or YAML file that specifies the resource to create.
    :param kubernetes_version:
        Version of the kubernetes API to use when creating the Resource. If
        omitted, the `latest` version will be used by default. Accepts either
        a string version label of a KubernetesVersion object.
    """
    with open(file_path) as f:
        return from_dict(json.load(f), kubernetes_version)



import json
import textwrap
import typing

import yaml
from kubernetes.client.rest import ApiException

from kuber import definitions

CREATE_SYMBOLS = {"AlreadyExists": "=", "NotFound": "!!"}

STATUS_SYMBOLS = {"NotFound": " "}

DELETE_SYMBOLS = {"NotFound": " "}


class ResponseInfo(typing.NamedTuple):
    """Data structure for parsed ApiExceptions."""

    resource: "definitions.Resource"
    symbol: str
    reason: str
    message: typing.Optional[str] = None
    exception: typing.Optional[ApiException] = None
    status: typing.Optional["definitions.Definition"] = None


def _parse_api_exception(
    resource: "definitions.Resource", error: ApiException, symbols: dict = None
) -> ResponseInfo:
    """Returns the parsed body of the given error"""
    try:
        body = json.loads(error.body)
    except json.JSONDecodeError:
        body = {}

    reason = body.get("reason", "UnknownError")
    return ResponseInfo(
        resource=resource,
        symbol=(symbols or {}).get(reason, "!!"),
        reason=reason,
        message="{}".format(body.get("message", error)),
        exception=error,
    )


def _echo_response(response: ResponseInfo):
    """Displays the response for human consumption."""
    resource_id = "{}/{}".format(
        response.resource.kind,
        response.resource.metadata.name or response.resource.kuber_uid,
    )
    wrap = "\n" if response.symbol == "!!" else ""
    print(f"{wrap}[{response.symbol}] {response.reason}: {resource_id}{wrap}")

    if response.symbol not in ("!!", "*"):
        return

    message = "\n".join(textwrap.wrap(response.message or "", 70))
    print(textwrap.indent(message, prefix="   "), end="\n\n")

    if response.symbol == "!!":
        print(textwrap.indent(response.resource.to_yaml(), prefix="   "), end="\n\n")


def create_resource(
    resource: "definitions.Resource", namespace: str = None, echo: bool = False
) -> ResponseInfo:
    """..."""
    try:
        status = resource.create_resource(namespace)
        response = ResponseInfo(resource, "+", "Created", status=status)
    except ApiException as error:
        response = _parse_api_exception(resource, error)

    if response.reason == "AlreadyExists":
        return patch_resource(resource, namespace, echo)

    if echo:
        _echo_response(response)
    return response


def replace_resource(
    resource: "definitions.Resource", namespace: str = None, echo: bool = False
) -> ResponseInfo:
    """..."""
    try:
        status = resource.replace_resource(namespace)
        response = ResponseInfo(resource, "-/+", "Replaced", status=status)
    except ApiException as error:
        response = _parse_api_exception(resource, error)

    if echo:
        _echo_response(response)
    return response


def patch_resource(
    resource: "definitions.Resource", namespace: str = None, echo: bool = False
) -> ResponseInfo:
    """..."""
    try:
        status = resource.patch_resource(namespace)
        response = ResponseInfo(resource, "~", "Updated", status=status)
    except ApiException as error:
        response = _parse_api_exception(resource, error)

    if echo:
        _echo_response(response)
    return response


def get_resource_status(
    resource: "definitions.Resource", namespace: str = None, echo: bool = False
) -> ResponseInfo:
    """Shows statuses of resources in the bundle."""
    try:
        status: definitions.Definition = resource.get_resource_status(
            namespace=namespace
        )
        if status is not None and status.to_dict():
            message = yaml.dump(status.to_dict(), default_flow_style=False)
            response = ResponseInfo(
                resource=resource,
                symbol="*",
                reason="Status",
                message=message,
                status=status,
            )
        else:
            status = resource.read_resource(namespace=namespace)
            response = ResponseInfo(
                resource=resource,
                symbol="*",
                reason="Status",
                message="Exists in cluster and is behaving as expected.",
                status=status,
            )
    except ApiException as error:
        response = _parse_api_exception(resource, error)

    if echo:
        _echo_response(response)
    return response


def delete_resource(
    resource: "definitions.Resource", namespace: str = None, echo: bool = False
) -> ResponseInfo:
    """..."""
    try:
        status = resource.delete_resource(namespace)
        response = ResponseInfo(resource, "-", "Deleted", status=status)
    except ApiException as error:
        response = _parse_api_exception(resource, error)

    if echo:
        _echo_response(response)
    return response
